@@ -12,3 +12,9 @@ from .tester_samples import (SRAMTester, InvTester, BufTester,
 from .random import random_bit, random_bv
 from .util import clog2
 from .spice_target import A2DError
+
+from fault.property import (assert_, implies, delay, posedge, repeat, goto,
+                            sequence, eventually, onehot0, onehot, countones,
+                            isunknown, past, rose, fell, stable, not_,
+                            throughout, until, until_with, inside)
+from fault.sva import sva
