@@ -61,7 +61,7 @@ class SystemVerilogTarget(VerilogTarget):
                  disp_type='on_error', waveform_file=None, coverage=False,
                  use_kratos=False, use_sva=False, skip_run=False,
                  no_top_module=False, vivado_use_system_verilog=True,
-                 disable_ndarray=False):
+                 disable_ndarray=False, fsdb_dumpvars_args=""):
         """
         circuit: a magma circuit
 
@@ -160,6 +160,9 @@ class SystemVerilogTarget(VerilogTarget):
                          Default is False except when the simulator is
                          "iverilog" when it is always True (since iverilog does
                          not currently support unpacked arrays)
+
+        fsdb_dumpvars_args: (optional) arguments to the `fsdbDumpvars()`
+                            function
         """
         # set default for list of external sources
         if include_verilog_libraries is None:
@@ -204,7 +207,7 @@ class SystemVerilogTarget(VerilogTarget):
         if simulator is None:
             raise ValueError("Must specify simulator when using system-verilog"
                              " target")
-        if simulator not in {"vcs", "ncsim", "iverilog", "vivado"}:
+        if simulator not in {"vcs", "ncsim", "xcelium", "iverilog", "vivado"}:
             raise ValueError(f"Unsupported simulator {simulator}")
 
         # save settings
@@ -247,7 +250,7 @@ class SystemVerilogTarget(VerilogTarget):
                                          "vcs waveform type")
                     suffix = self.waveform_type
                 self.waveform_file = f"waveforms.{self.waveform_type}"
-            elif self.simulator in {"ncsim", "iverilog", "vivado"}:
+            elif self.simulator in {"ncsim", "xcelium", "iverilog", "vivado"}:
                 self.waveform_file = "waveforms.vcd"
             else:
                 raise NotImplementedError(self.simulator)
@@ -264,6 +267,11 @@ class SystemVerilogTarget(VerilogTarget):
                 raise ImportError("Cannot find kratos-runtime in the system. "
                                   "Please do \"pip install kratos-runtime\" "
                                   "to install.")
+        self.fsdb_dumpvars_args = fsdb_dumpvars_args
+
+        # set up cadence tools command
+        self.ncsim_cmd = self.cadence_cmd("irun")
+        self.xcelium_cmd = self.cadence_cmd("xrun")
 
     def add_decl(self, type_, name, exist_ok=False):
         if str(name) in self.declarations:
@@ -382,17 +390,20 @@ class SystemVerilogTarget(VerilogTarget):
     def make_delay(self, i, action):
         return [f'#({action.time}*1s);']
 
-    def make_print(self, i, action):
-        # build up argument list for the $write command
+    def _make_print_args(self, values):
         args = []
-        args.append(f'"{action.format_str}"')
-        for port in action.ports:
+        for port in values:
             if isinstance(port, (Number, AbstractBit, AbstractBitVector)) and \
                     not isinstance(port, m.Bits):
                 args.append(f'{port}')
             else:
                 args.append(f'{self.make_name(port)}')
+        return args
 
+    def make_print(self, i, action):
+        # build up argument list for the $write command
+        args = [f'"{action.format_str}"']
+        args += self._make_print_args(action.ports)
         # generate the command
         args = ', '.join(args)
         return [f'$write({args});']
@@ -534,6 +545,13 @@ class SystemVerilogTarget(VerilogTarget):
                     cond = f'({name} == {value})'
                 err_msg = 'Expected %x, got %x'
                 err_args = [value, name]
+        if action.msg is not None:
+            if isinstance(action.msg, str):
+                err_msg += "\\n" + action.msg
+            else:
+                assert isinstance(action.msg, tuple)
+                err_msg += "\\n" + action.msg[0]
+                err_args += self._make_print_args(action.msg[1:])
 
         # construct the body of the $error call
         err_fmt_str = f'"{err_hdr}.  {err_msg}."'
@@ -648,7 +666,7 @@ class SystemVerilogTarget(VerilogTarget):
                                  f'$vcdplusmemon();']
             if self.waveform_type == "fsdb":
                 initial_body += [f'$fsdbDumpfile("{self.waveform_file}");',
-                                 f'$fsdbDumpvars();']
+                                 f'$fsdbDumpvars({self.fsdb_dumpvars_args});']
         elif self.dump_waveforms and self.simulator in {"iverilog", "vivado"}:
             # https://iverilog.fandom.com/wiki/GTKWAVE
             initial_body += [f'$dumpfile("{self.waveform_file}");',
@@ -719,14 +737,22 @@ class SystemVerilogTarget(VerilogTarget):
         vlog_srcs += self.include_verilog_libraries
 
         # generate simulator commands
-        if self.simulator == 'ncsim':
+        if self.simulator == 'ncsim' or self.simulator == "incisive":
             # Compile and run simulation
-            cmd_file = self.write_ncsim_tcl()
+            cmd_file = self.write_cadence_tcl()
             sim_cmd = self.ncsim_cmd(sources=vlog_srcs, cmd_file=cmd_file)
             sim_err_str = None
             # Skip "bin_cmd"
             bin_cmd = None
             bin_err_str = None
+        elif self.simulator == 'xcelium':
+            # Compile and run simulation
+            cmd_file = self.write_cadence_tcl()
+            sim_cmd = self.xcelium_cmd(sources=vlog_srcs, cmd_file=cmd_file)
+            sim_err_str = None
+            # Skip "bin_cmd"
+            bin_cmd = None
+            bin_err_str = None
         elif self.simulator == 'vivado':
             # Compile and run simulation
             cmd_file = self.write_vivado_tcl(sources=vlog_srcs)
@@ -815,7 +841,7 @@ class SystemVerilogTarget(VerilogTarget):
         # also add the directory to the current LD_LIBRARY_PATH
         self.sim_env["LD_LIBRARY_PATH"] = os.path.abspath(self.directory)
 
-    def write_ncsim_tcl(self):
+    def write_cadence_tcl(self):
         # construct the TCL commands to run the Incisive/Xcelium simulation
         tcl_cmds = []
         if self.dump_waveforms:
@@ -909,57 +935,59 @@ class SystemVerilogTarget(VerilogTarget):
             retval += [def_arg]
         return retval
 
-    def ncsim_cmd(self, sources, cmd_file):
-        cmd = []
+    def cadence_cmd(self, tool_name):
+        def cmd_fn(sources, cmd_file):
+            cmd = []
 
-        # binary name
-        cmd += ['irun']
+            # binary name
+            cmd += [tool_name]
 
-        # add any extra flags
-        cmd += self.flags
+            # add any extra flags
+            cmd += self.flags
 
-        # send name of top module to the simulator
-        if not self.no_top_module:
-            cmd += ['-top', f'{self.top_module}']
+            # send name of top module to the simulator
+            if not self.no_top_module:
+                cmd += ['-top', f'{self.top_module}']
 
-        # timescale
-        cmd += ['-timescale', f'{self.timescale}']
+            # timescale
+            cmd += ['-timescale', f'{self.timescale}']
 
-        # TCL commands
-        cmd += ['-input', f'{cmd_file}']
+            # TCL commands
+            cmd += ['-input', f'{cmd_file}']
 
-        # source files
-        cmd += [f'{src}' for src in sources]
+            # source files
+            cmd += [f'{src}' for src in sources]
 
-        # library files
-        for lib in self.ext_libs:
-            cmd += ['-v', f'{lib}']
+            # library files
+            for lib in self.ext_libs:
+                cmd += ['-v', f'{lib}']
 
-        # include directory search path
-        for dir_ in self.inc_dirs:
-            cmd += ['-incdir', f'{dir_}']
+            # include directory search path
+            for dir_ in self.inc_dirs:
+                cmd += ['-incdir', f'{dir_}']
 
-        # define variables
-        cmd += self.def_args(prefix='+define+')
+            # define variables
+            cmd += self.def_args(prefix='+define+')
 
-        # misc flags
-        if self.dump_waveforms:
-            cmd += ["-access", "r"]
-        cmd += ['-notimingchecks']
-        if self.no_warning:
-            cmd += ['-neverwarn']
+            # misc flags
+            if self.dump_waveforms:
+                cmd += ["-access", "r"]
+            cmd += ['-notimingchecks']
+            if self.no_warning:
+                cmd += ['-neverwarn']
 
-        # kratos flags
-        if self.use_kratos:
-            from kratos_runtime import get_ncsim_flag
-            cmd += get_ncsim_flag().split()
+            # kratos flags
+            if self.use_kratos:
+                from kratos_runtime import get_ncsim_flag
+                cmd += get_ncsim_flag().split()
 
-        # coverage flags
-        if self.coverage:
-            cmd += ["-coverage", "b", "-covoverwrite"]
+            # coverage flags
+            if self.coverage:
+                cmd += ["-coverage", "b", "-covoverwrite"]
 
-        # return arg list
-        return cmd
+            # return arg list
+            return cmd
+        return cmd_fn
 
     def vivado_cmd(self, cmd_file):
         cmd = []
